<?php

/**
 * The CalendarObject represents a single VEVENT or VTODO within a Calendar.
 *
 * @package Sabre
 * @subpackage CalDAV
 * @copyright Copyright (C) 2007-2012 Rooftop Solutions. All rights reserved.
 * @author Evert Pot (http://www.rooftopsolutions.nl/) 
 * @license http://code.google.com/p/sabredav/wiki/License Modified BSD License
 */
class Sabre_CalDAV_CalendarObject extends Sabre_DAV_File implements Sabre_CalDAV_ICalendarObject, Sabre_DAVACL_IACL {

    /**
     * Sabre_CalDAV_Backend_Abstract
     *
     * @var array
     */
    protected $caldavBackend;

    /**
     * Array with information about this CalendarObject
     *
     * @var array
     */
    protected $objectData;

    /**
     * Array with information about the containing calendar
     *
     * @var array
     */
    protected $calendarInfo;

    /**
     * Constructor
     *
     * @param Sabre_CalDAV_Backend_Abstract $caldavBackend
     * @param array $calendarInfo
     * @param array $objectData
     */
    public function __construct(Sabre_CalDAV_Backend_Abstract $caldavBackend,array $calendarInfo,array $objectData) {

        $this->caldavBackend = $caldavBackend;

        if (!isset($objectData['calendarid'])) {
            throw new InvalidArgumentException('The objectData argument must contain a \'calendarid\' property');
        }
        if (!isset($objectData['uri'])) {
            throw new InvalidArgumentException('The objectData argument must contain an \'uri\' property');
        }

        $this->calendarInfo = $calendarInfo;
        $this->objectData = $objectData;

    }

    /**
     * Returns the uri for this object
     *
     * @return string
     */
    public function getName() {

        return $this->objectData['uri'];

    }

    /**
     * Returns the ICalendar-formatted object
     *
     * @return string
     */
    public function get() {

        // Pre-populating the 'calendardata' is optional, if we don't have it
        // already we fetch it from the backend.
        if (!isset($this->objectData['calendardata'])) {
            $this->objectData = $this->caldavBackend->getCalendarObject($this->objectData['calendarid'], $this->objectData['uri']);
        }
        if(array_key_exists('shareduri',$this->calendarInfo) && $this->calendarInfo['permission']=='busy'){
//            dbg("WASH for FREEBUSY ==>".vl($this->objectData));
            //$this->objectData['calendardata'] = 
            $new_ics = $this->washFreeBusy($this->objectData['calendardata'], $this->calendarInfo['shared_by']);
//            dbg("WASHED for FREEBUSY ==>".vl($new_ics));
            $this->objectData['calendardata'] = $new_ics;
        }
//        dbg("***ICS ".vl($this->objectData));

        return $this->objectData['calendardata'];

    }

    /**
     * Updates the ICalendar-formatted object
     *
     * @param string $calendarData
     * @return void
     */
    public function put($calendarData) {

        if (is_resource($calendarData)) {
            $calendarData = stream_get_contents($calendarData);
        }
        $etag = $this->caldavBackend->updateCalendarObject($this->calendarInfo['id'],$this->objectData['uri'],$calendarData);
        $this->objectData['calendardata'] = $calendarData;
        $this->objectData['etag'] = $etag;

        return $etag;

    }

    /**
     * Deletes the calendar object
     *
     * @return void
     */
    public function delete() {

        $this->caldavBackend->deleteCalendarObject($this->calendarInfo['id'],$this->objectData['uri']);

    }

    /**
     * Returns the mime content-type
     *
     * @return string
     */
    public function getContentType() {

        return 'text/calendar';

    }

    /**
     * Returns an ETag for this object.
     *
     * The ETag is an arbitrary string, but MUST be surrounded by double-quotes.
     *
     * @return string
     */
    public function getETag() {

        if (isset($this->objectData['etag'])) {
            return $this->objectData['etag'];
        } else {
            return '"' . md5($this->get()). '"';
        }

    }

    /**
     * Returns the last modification date as a unix timestamp
     *
     * @return time
     */
    public function getLastModified() {

        return $this->objectData['lastmodified'];

    }

    /**
     * Returns the size of this object in bytes
     *
     * @return int
     */
    public function getSize() {

        if (array_key_exists('size',$this->objectData)) {
            return $this->objectData['size'];
        } else {
            return strlen($this->get());
        }

    }

    /**
     * Returns the owner principal
     *
     * This must be a url to a principal, or null if there's no owner
     *
     * @return string|null
     */
    public function getOwner() {

        return $this->calendarInfo['principaluri'];

    }

    /**
     * Returns a group principal
     *
     * This must be a url to a principal, or null if there's no owner
     *
     * @return string|null
     */
    public function getGroup() {

        return null;

    }

    /**
     * Returns a list of ACE's for this node.
     *
     * Each ACE has the following properties:
     *   * 'privilege', a string such as {DAV:}read or {DAV:}write. These are
     *     currently the only supported privileges
     *   * 'principal', a url to the principal who owns the node
     *   * 'protected' (optional), indicating that this ACE is not allowed to
     *      be updated.
     *
     * @return array
     */
    public function getACL() {
        $uri='principaluri';
        
        $read = true; $write=true; $shared = true;
        if(array_key_exists('shareduri',$this->calendarInfo)){
            $uri='shareduri';
            $shared=true;
//            dbg("ACL ICS getACL uri=".vl($this->calendarInfo['uri'])." shareduri=".$this->calendarInfo['shareduri']);
            if($this->calendarInfo['permission']!='write')
                $write=false;
        }else{
//            dbg("ACL ICS NOT getACL shareduri [".vl($this->calendarInfo['uri'])."]");
        }
        $ret = array();
 
        if($read){
            $ret[] = array(
                'privilege' => '{DAV:}read',
                'principal' => $this->calendarInfo[$uri],
                'protected' => true,
            );
        }
        if($write){
            $ret[] = array(
                'privilege' => '{DAV:}write',
                'principal' => $this->calendarInfo[$uri],
                'protected' => true,
            );
       }
       if($shared){
            $ret[] = array(
                'privilege' => '{DAV:}read',
                'principal' => $this->calendarInfo[$uri] . '/calendar-proxy-write',
                'protected' => true,
            );
            $ret[] = array(
                'privilege' => '{DAV:}write',
                'principal' => $this->calendarInfo[$uri] . '/calendar-proxy-write',
                'protected' => true,
            );
            $ret[] = array(
                'privilege' => '{DAV:}read',
                'principal' => $this->calendarInfo[$uri] . '/calendar-proxy-read',
                'protected' => true,
            );
        }
        return $ret;


    }

    /**
     * Updates the ACL
     *
     * This method will receive a list of new ACE's.
     *
     * @param array $acl
     * @return void
     */
    public function setACL(array $acl) {

        throw new Sabre_DAV_Exception_MethodNotAllowed('Changing ACL is not yet supported');

    }

    /**
     * Returns the list of supported privileges for this node.
     *
     * The returned data structure is a list of nested privileges.
     * See Sabre_DAVACL_Plugin::getDefaultSupportedPrivilegeSet for a simple
     * standard structure.
     *
     * If null is returned from this method, the default privilege set is used,
     * which is fine for most common usecases.
     *
     * @return array|null
     */
    public function getSupportedPrivilegeSet() {

        return null;

    }

    /**
     * Returns the path to real principal
     *
     * @return string
     */
    public function sharedPath($path) {

        if(!array_key_exists('shareduri',$this->calendarInfo)) return $path;
        
        $xpath = str_replace( $this->calendarInfo['shared_by'], $this->calendarInfo['shared_with'], $path );

        dbg("*** UPDATED 1 [$path]->[$xpath]");

        return $xpath;
    }


    public function washFreeBusy($ics, $sharing)
    {
        dbg("WASHING");
        
        $pp = preg_split("/[\/@]+/",$sharing);
        $name = $pp[0];
 
        $lines = explode("\n",$ics);
        $updated = array();
        foreach($lines as $line){
            if(substr($line,0,8)=='SUMMARY:') {$updated[]="SUMMARY: $name busy"; continue;}
            if(substr($line,0,9)=='LOCATION:') {$updated[]="LOCATION:"; continue;}
            if(substr($line,0,12)=='DESCRIPTION:') {$updated[]="DESCRIPTION:"; continue;}
            
            $updated[]=$line;
        }
        $new_ics=implode("\n",$updated);
        dbg("WASHED ".vl($new_ics));
        return $new_ics;
    }

    
}

